import logging
from typing import Dict, Union

from easybits_module_interface.models import Message, MessageTypes

logger = logging.getLogger()


def with_error_response(error_message: Union[str, Dict], exception_type) -> Message:
    """
    Decorator to handle exceptions and return an error message.

    :param error_message: The error message to return.
    :type error_message: str | Dict
    :param exception_type: The exception type to catch.
    :type exception_type: Exception
    :return: The error message.


    Example:
        >>> from easybits_module_interface.utils import with_error_response
        >>> from easybits_module_interface.models import Message
        >>> @with_error_response('An error occurred: {error}', Exception)
        >>> def process_message(message: Message) -> Message:
        ...     raise Exception('Something went wrong')
        >>> message = Message()
        >>> message.message.incoming = 'Hello'
        >>> message.meta.language = 'en'
        >>> message = process_message(message)
        >>> message.message.outgoing
        'An error occurred: Something went wrong'
    """
    def decorator(func):
        def wrapper(message: Message, *args, **kwargs) -> Message:
            """
            Wrapper function to handle exceptions and return an error message.

            :param message: The original message object.
            :param args: The arguments.
            :param kwargs: The keyword arguments.
            :return: The error message.
            """
            try:
                return func(message, *args, **kwargs)
            except exception_type as ex:
                logger.error(f'Error while processing message: {ex}')
                em = error_message
                if isinstance(em, dict):
                    em = em.get(message.meta.language, 'en')
                message.message.outgoing = em.replace('{error}', str(ex))
                message.meta.type = MessageTypes.OUTGOING_MESSAGE.value
                return message
        return wrapper
    return decorator


def with_error_response_cls(error_message: Union[str, Dict], exception_type) -> Message:
    """
    Decorator to handle exceptions and return an error message.

    **Note:** This decorator is used for class methods.

    :param error_message: The error message to return.
    :type error_message: str | Dict
    :param exception_type: The exception type to catch.
    :type exception_type: Exception
    :return: The error message.


    Example:
        >>> from easybits_module_interface.utils import with_error_response_cls
        >>> from easybits_module_interface.models import Message
        >>> class MyClass:
        ...     @with_error_response_cls('An error occurred: {error}', Exception)
        ...     def process_message(self, message: Message) -> Message:
        ...         raise Exception('Something went wrong')
        >>> message = Message()
        >>> message.message.incoming = 'Hello'
        >>> message.meta.language = 'en'
        >>> message = MyClass().process_message(message)
        >>> message.message.outgoing
        'An error occurred: Something went wrong'
    """
    def decorator(func):
        def wrapper(self, message: Message, *args, **kwargs) -> Message:
            """
            Wrapper function to handle exceptions and return an error message.

            :param self: The class instance.
            :param message: The original message object.
            :param args: The arguments.
            :param kwargs: The keyword arguments.
            :return: The error message.
            """
            try:
                return func(self, message, *args, **kwargs)
            except exception_type as ex:
                logger.error(f'Error while processing message: {ex}')
                em = error_message
                if isinstance(em, dict):
                    em = em.get(message.meta.language, 'en')
                message.message.outgoing = em.replace('{error}', str(ex))
                message.meta.type = MessageTypes.OUTGOING_MESSAGE.value
                return message
        return wrapper
    return decorator



from ._internal import ModuleInterface, main, bootstrap_module_class  # noqa
from ._version import __version__  # noqa
from .task_execution.task_executor import TaskExecutor
from .task_execution.models import GenerationRequest
from .state_machine import (IllegalTransitionError, StateMachine,  # noqa
                            Transition)

__all__ = [
    '__version__',
    'main',
    'boostrap_module_class',
    'ModuleInterface',
    'TaskExecutor',
    'GenerationRequest',
    'StateMachine',
    'Transition',
    'IllegalTransitionError',
]



import asyncio
from dataclasses import dataclass, field, asdict
from functools import partial
from datetime import datetime
from itertools import count
from typing import Callable, Optional
from collections import defaultdict
from easybits_module_interface.models import Message, MessageTypes
from uuid import uuid4
from copy import deepcopy
from easybits_module_interface.task_execution.thread_pool_executor import ManagedThreadPoolExecutor
from easybits_module_interface.task_execution.models import GenerationRequest, RunningTask


def sync_wrapper(func: Callable) -> Callable:
    """
    Decorator style implementation wrapping any callable Future
    into a synchronous task.
    """
    def inner(*args, **kwargs):
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        result = loop.run_until_complete(func(*args, **kwargs))
        loop.close()
        return result
    return inner


class TaskExecutor:
    """
    TaskExecutor class to manage the execution of tasks in a thread pool.

    The class is responsible for:
    - Scheduling tasks in a thread pool
    - Handling finished tasks
    - Notifying users about timeouts
    - Sending out results to users
    - Sending out verbose results to users
    - Publishing status messages
    - Handling rate limits
    - Handling task timeouts
    - Handling errors

    To add custom behavior for...
    - Notifying users about timeouts: Override the `notify_timeout` method
    - Sending out verbose results: Override the `notify_verbose_result` method
    - Notifying users about task timeouts: Override the `notify_execution_timeout` method

    """
    def __init__(self, error_responses, logger):
        # token based process map, to maintain pending tasks
        self.process_queue = {}
        # status messages used to sed out status messages from running/pending tasks
        self.status_messages = []
        # copy of/reference to error responses dictionary from module
        self.error_responses = error_responses
        # constant for token based rate limiting
        self.max_tasks_per_owner = 5
        # constant for number of parallel threads (keep it 64!)
        self.max_running_tasks = 64
        # map for detailed rate limits - currently everyone is treaded equally
        self.max_running_tasks_owner_map = defaultdict(lambda: self.max_tasks_per_owner)
        # list of currently running tasks
        self.running_tasks = []
        # map for rate limits
        self.running_tasks_map = defaultdict(int)
        # a reference to the module logger
        self.logger = logger

    async def publish(self, message: Message) -> None:
        """
        Method to publish a message to RMQ.
        Dynamically set this to inject outside functionality.

        See the module_implementation.py for more details.
        """
        raise NotImplementedError('Attach method after instantiation!')

    async def notify_timeout(self, process: GenerationRequest) -> None:
        """
        Method to notify a user about a timeout in their task.

        Override this method to implement custom behavior.
        """
        pass

    def _update_rate_limit(self, owner_id, change: int) -> None:
        self.running_tasks_map[owner_id] += change
        if self.running_tasks_map[owner_id] <= 0:
            del self.running_tasks_map[owner_id]

    async def notify(self, processes: list[GenerationRequest | RunningTask]) -> None:
        """
        Method to notify users about progress in processes.

        :param processes: List of running task objects
        :returns: Nothing
        """
        for p in processes:
            if isinstance(p, GenerationRequest):
                process = p
            elif isinstance(p, RunningTask):
                process = p.generation_request
            else:
                raise ValueError('Invalid type for process')

            # notify unnotified users about delays in their tasks
            if not process.notified and (datetime.now() - process.timestamp).seconds > process.timeout:
                await self.notify_timeout(process)
                process.notified = True

    async def schedule_task(self, free_workers: int, pool: ManagedThreadPoolExecutor) -> None:
        """
        Method to schedule a pending task in the thread pool.

        1. iterates over list of pending tasks
        2. checks for rate limit
        3. then schedules the first task that is allowed to be scheduled

        :param free_workers: number of free workers in the thread pool
        :param pool: the pool to schedule the task in
        :returns: Nothing
        """
        if min(free_workers, len(self.process_queue)) <= 0:
            return

        found = False
        p_id = None
        # search for task that is allowed to be scheduled
        for p_id, p in self.process_queue.items():
            running_for_owner = self.max_running_tasks_owner_map[p.owner_id]
            if (p.owner_id not in self.running_tasks_map
                or self.running_tasks_map[p.owner_id] < running_for_owner):
                found = True
                break

        if not found:
            return

        # schedule task
        generation_request = self.process_queue.pop(p_id)
        generation_request.execution_timestamp = datetime.now()
        thread = pool.submit(sync_wrapper(generation_request.method), p_id)
        task = RunningTask(generation_request, thread)
        self.running_tasks.append(task)

        # update rate limit map
        self._update_rate_limit(generation_request.owner_id, 1)

        self.logger.debug(
            f'Scheduled task {generation_request.task_id}. '
            f'{len(self.running_tasks)} tasks running. '
            f'{len(self.process_queue)} tasks pending. '
            f'{len(pool.task_map)} executors in pool running '
            f'{sum(len(v) for v in pool.task_map.values())} tasks ('
            f'{len(pool.frozen_executors)} frozen executors with '
            f'{sum(len(v) for v in pool.frozen_tasks.values())} tasks)'
        )

    async def notify_verbose_result(self, msg: Message) -> None:
        """
        Method to notify the user about a verbose result.

        Override this method to implement custom behavior.
        """
        pass

    async def sendout_result(self, msg: Message, is_success: bool) -> None:
        """
        Send out the result of a task to the user.

        :param msg: The original message to send out
        :param is_success: Whether the task was successfully executed
        :return: None
        """
        msg.meta.event = 'success_response' if is_success else 'error_response'
        await self.publish(msg)
        verbose = msg.config.get('verbose', 0)
        try:
            verbose = int(verbose)
        except ValueError:
            verbose = 0

        if not is_success or verbose < 2:
            return

        # sendout verbose message
        await self.notify_verbose_result(msg)

    async def handle_finished_task(self, running_task: GenerationRequest) -> None:
        """
        Method to handle a finished task

        :param running_task: The reference to the finished task
        :returns: Nothing
        """
        result = running_task.task.result()
        if result:
            message, is_success = result
            await self.sendout_result(message, is_success)

        self._update_rate_limit(running_task.generation_request.owner_id, -1)
        self.running_tasks_map[running_task.generation_request.owner_id] -= 1
        self.logger.debug(
            f'Task {running_task.generation_request.task_id} done. '
            f'{len(self.running_tasks) - 1} tasks running. '
            f'{len(self.process_queue)} tasks pending. '
        )

    async def notify_execution_timeout(self, running_task: GenerationRequest) -> None:
        """
        Method to notify a user about a timeout in their task.

        Override this method to implement custom behavior.

        :param running_task: The reference to the task that timed out
        """
        pass

    async def handle_execution_timeout(self, running_task: GenerationRequest) -> None:
        """
        Method to handle a task that timed out.

        :param running_task: The reference to the task that timed out
        :returns: Nothing
        """
        success = running_task.task.cancel()
        self._update_rate_limit(running_task.generation_request.owner_id, -1)

        if running_task.generation_request.verbose > 1:
            # send out verbose message for timeout
            await self.notify_execution_timeout(running_task)

        self.logger.debug(
            f'TIMEOUT: Task {running_task.generation_request.task_id} killed. '
            f'{len(self.running_tasks) - 1} tasks running. '
            f'{len(self.process_queue)} tasks pending. '
        )

    async def check_finished_tasks(self) -> tuple[list[int], list[int]]:
        """
        Check for finished (positive/negative) tasks and handle them appropriately

        :returns: Tuple of list of still pending tasks, successful tasks and list of cancelled tasks
        """
        pending_tasks = []
        cancelled_tasks = []
        successful_tasks = []
        for running_task in self.running_tasks:
            idx = running_task.generation_request.task_id
            if running_task.task.done():
                await self.handle_finished_task(running_task)
                successful_tasks.append(idx)
                continue
            elif ((datetime.now() - running_task.generation_request.execution_timestamp).seconds 
                > running_task.generation_request.execution_timeout * 60):
                await self.handle_execution_timeout(running_task)
                cancelled_tasks.append(idx)
                continue
            else:
                pending_tasks.append(idx)
        return pending_tasks, successful_tasks, cancelled_tasks

    async def publish_status_messages(self) -> None:
        """
        Helper method to publish scheduled status messages

        :returns: Nothing
        """
        if self.status_messages:
            for msg in self.status_messages:
                await self.publish(msg)
            self.status_messages = []

    async def run(self) -> None:
        """
        Consume scheduled tasks from the `process_queue` and execute them.

        The tasks are executed in a thread pool, so the main event loop is not blocked.
        While processing, the module will occasional publish status messages to the `outgoing_messages` exchange.

        If threads become stale due to issues on HF the thread pool will notify and terminate the current
        process.
        This will cause a call of this method again.
        """
        pool = ManagedThreadPoolExecutor(None, self.logger)
        while True:
            if not self.process_queue and not self.running_tasks and not self.status_messages:
                if pool.requires_restart:
                    self.logger.debug('Restarting task executor')
                    break
                await asyncio.sleep(.1)
                continue

            # schedule new task
            free_workers = self.max_running_tasks - len(self.running_tasks)
            await self.schedule_task(free_workers, pool)

            # notify user about timeouts in pending tasks
            await self.notify(list(self.process_queue.values()))

            # check for finished tasks
            pending_tasks, successful_tasks, cancelled_tasks = await self.check_finished_tasks()
            self.running_tasks = [t for t in filter(lambda x: x.generation_request.task_id in pending_tasks, self.running_tasks)]

            # notify user about timeouts in running tasks
            await self.notify(self.running_tasks)

            # publish status messages produced by tasks
            await self.publish_status_messages()
            pool.garbage_collect(successful_tasks, cancelled_tasks)
            await asyncio.sleep(1)
        pool.shutdown()




from copy import deepcopy
from dataclasses import dataclass, field, fields, make_dataclass
from enum import Enum
from typing import Dict, List, Optional, Union


class BaseDataClass:
    """
    Base class for dataclasses with a from_dict method.
    """
    @classmethod
    def get_model_kwargs(cls, kwargs: Dict) -> Dict:
        """
        Getter for kwargs based on the fields of the class.

        :param kwargs: kwargs to filter
        :return: filtered kwargs
        """
        field_names = [field.name for field in fields(cls)]
        kwargs_out = deepcopy(kwargs)
        for f in kwargs.keys():
            if f not in field_names:
                kwargs_out.pop(f, None)
        return kwargs_out

    @classmethod
    def from_dict(cls, **kwargs):
        """
        Safe method to create an instance from given kwargs.

        :param kwargs: object as kwargs
        :return: instance of the class
        """
        return cls(**cls.get_model_kwargs(kwargs))


class DynamicFieldsDataClass(BaseDataClass):
    """
    Base class for dataclasses with dynamic fields.

    Example:
        >>> @dataclass
        ... class MyFoo(DynamicFieldsDataClass):
        ...     foo: str
        ...
        >>> obj = MyFoo.from_dict(foo='string', bar=1)
        >>> obj.foo
        'string'
        >>> obj.bar
        1
        >>> asdict(obj)
        {'foo': 'string', 'bar': 1}
        >>> MyFoo(foo='string', bar=1)
        AttributeError: 'MyFoo' object has no attribute 'bar'
    """
    @classmethod
    def from_dict(cls, **kwargs):
        """
        Creates a new dataclass with the fields provided in the kwargs.

        New fields are added to the class definition as Optional[str].

        :param kwargs: object as kwargs
        :return: new dataclass instance
        """
        field_names = [field.name for field in fields(cls)]
        new_fields = [(k, Optional[str], None) for k, v in kwargs.items() if k not in field_names]
        dynamic_cls = make_dataclass(cls.__name__, new_fields, bases=(cls,))
        return dynamic_cls(**kwargs)


class MessageTypes(Enum):
    SYSTEM_MESSAGE_PARSER = 'system_message_parser'
    SYSTEM_MESSAGE_CONTROLLER = 'system_message_controller'
    SYSTEM_MESSAGE_DATASINK = 'system_message_datasink'
    INCOMING_MESSAGE = 'incoming_message'
    OUTGOING_MESSAGE = 'outgoing_message'
    OUTGOING_MESSAGE_IMAGE = 'outgoing_message_image'
    OUTGOING_MESSAGE_AUDIO = 'outgoing_message_audio'


class SystemMessageEvents(Enum):
    CONFIGURE_MODULE = 'configure_module'
    SAVE = 'save'
    DELETE = 'delete'
    CONFIGURE_PROXY_MODE = 'configure_proxy_mode'
    # deprecated, kept for backward compatibility
    UPDATE_COLLECTION = 'update_collection'
    UPDATE_COLLECTION_DATA = 'update_collection_data'


@dataclass
class MessageMeta(DynamicFieldsDataClass):
    """
    Meta data for messages.

    This class is dynamic, meaning using the from_dict method
    will create a new dataclass with the fields provided in the kwargs.

    This is useful to add more meta data to the message without
    changing the class definition.
    """
    #: type of message
    type: MessageTypes
    #: event type of message
    event: Optional[str] = None
    #: internal ID of communication channel
    communicationChannelId: Optional[str] = None
    #: type identifier of the controller used for communication
    controller: Optional[str] = None
    #: ID of the user can be used for tracking/state management
    internalUserId: Optional[str] = None
    #: ID of the bot used for communication
    botId: Optional[str] = None

    #: [Optional] configured language of bot used inside modules to generate responses
    language: Optional[str] = None
    #: [Optional] internal tracking ID of the message
    messageId: Optional[str] = None


@dataclass
class UserMessage(BaseDataClass):
    """
    Object to describe message content aka a UserMessage
    """
    #: content of incoming message
    incoming: Optional[str] = None
    #: content of incoming file
    incoming_file: Optional[str] = None
    #: content of outgoing message
    outgoing: Optional[str] = None


@dataclass
class ContextProvider(BaseDataClass):
    """
    Configuration for a context provider
    """
    #: name of resulting key in context
    key: str = ""
    #: name of the context provider to use
    provider_name: str = ""


@dataclass
class AvailableModule(BaseDataClass):
    """
    Module config template
    """
    #: Internal ID
    id: int
    #: verbose identifier of module
    name: str
    #: description of module
    description: str
    #: indicates that the module supports reengagement
    supports_reengagement: bool = False
    #: Layout of configuration values
    configuration_layout: Optional[Dict] = None
    #: List of supported context providers by the module
    supported_context_providers: Optional[List] = None
    #: Indicates if the module is allowed to be used
    #:   without any pre-selection process
    is_passive: bool = False
    # The following attributes are commented out, as they
    # are only used in the client applications.
    #: Indicates if the module is available to be used
    #is_published: bool = True
    #: Indicates if the module is visible for users
    #is_visible: bool = True
    #: Indicates if the module is the default module
    #    that is used during the default module prediction
    #    phase of the parser.
    is_default: bool = False
    #: String of comma separated intentions that the module supports
    intentions: Optional[str] = None


@dataclass
class BotModule(BaseDataClass):
    """
    Configuration for a bot specific module.
    """
    #: internal identifier of bot module
    id: int
    #: instance of :py:obj:`~.AvailableModule`
    module: AvailableModule = None
    #: indicates that the module is active
    is_active: bool = False
    #: indicates that the module uses reengagement
    has_reengagement: bool = False
    #: list of instances of :py:obj:`~.ContextProvider` \
    #  that are used to extract context from incoming messages
    context_providers: List[ContextProvider] = field(default_factory=list)
    #: configuration values for the module
    configuration_values: Dict = field(default_factory=dict)
    #: configuration values for the parser
    parser_configuration: Dict = field(default_factory=dict)
    #: desired reengagement message
    reengagement_message: str = None
    #: Timeout in minutes for reengagement message
    reengagement_timeout_minutes: int = None

    @property
    def name(self):
        return self.module.name

    @property
    def intentions(self):
        return self.module.intentions

    @classmethod
    def from_dict(cls, **kwargs):
        """
        Creates an instance from given kwargs

        :param kwargs: object as kwargs
        :return: :py:obj:`~.BotModule` instance
        """
        module = kwargs.pop('module', None)
        context_providers = kwargs.pop('context_providers', [])

        return super().from_dict(
            **kwargs,
            module=AvailableModule.from_dict(**module) if module else None,
            context_providers=[ContextProvider.from_dict(**c) for c in context_providers]
        )


@dataclass
class ConfigureModuleContext(BaseDataClass):
    """
    Integration config template
    """
    id: str = None
    #: name of the integration
    name: str = None
    #: list of instances of :py:obj:`~.ContextProvider`
    context_providers: List[ContextProvider] = field(default_factory=list)
    #: indicates that the integration is ready to consume messages
    is_active: bool = False
    #: indicates that the integration supports reengagement
    re_engagement: bool = False

    #: layout of configuration values
    configuration_layout: Optional[Dict] = None
    #: indicates that the module supports reengagement
    supports_reengagement: bool = False
    #: description of module
    description: str = None


@dataclass
class Message(BaseDataClass):
    #: message meta data
    meta: MessageMeta
    #: extracted context from incoming message
    context: Optional[Union[Dict, ConfigureModuleContext]] = None
    #: message content
    message: Optional[UserMessage] = None
    #: configuration values forwarded to the module
    config: Optional[Dict] = None

    @classmethod
    def from_dict(cls, **kwargs):
        """
        Creates an instance from given kwargs

        :param kwargs: object as kwargs
        :return: :py:obj:`~.Message` instance
        """
        meta_data = kwargs.pop('meta', None)
        if meta_data is not None:
            meta_data = MessageMeta.from_dict(**meta_data)
        message = kwargs.pop('message', None)
        if message is not None:
            message = UserMessage.from_dict(**message)

        context = kwargs.pop('context', None)
        is_configure_module_message = (
            meta_data
            and meta_data.type == MessageTypes.SYSTEM_MESSAGE_PARSER.value
            and meta_data.event in [
                SystemMessageEvents.SAVE.value,
                SystemMessageEvents.DELETE.value,
            ]
        )
        if is_configure_module_message:
            context = AvailableModule.from_dict(**context)

        config = kwargs.pop('config', None)

        return super().from_dict(
            context=context,
            meta=meta_data,
            message=message,
            config=config,
        )



import asyncio
import json
import sentry_sdk
import traceback
import logging
import logging.config
from dataclasses import asdict
from typing import Callable, Optional
from aio_pika import Message as PikaMessage, ExchangeType, connect_robust
from aio_pika.abc import AbstractIncomingMessage

from easybits_module_interface.models import Message


class ConnectionError(Exception):
    """
    Exception raised when connection to RabbitMQ fails.
    """
    pass


class ConnectionSetup:
    """
    Base class for classes that need to connect to RabbitMQ.

    Provides a :py:obj:`~connection` and :py:obj:`~channel` attribute.

    :param connection: Connection to RabbitMQ

    Example::
        >>> import aio_pika
        >>> connection = await aio_pika.connect("amqp://localhost/", loop=loop)
        >>> setup = ConnectionSetup(connection)
        >>> await setup.connect()
        >>> await setup.channel.declare_queue("test")
        >>> await setup.disconnect()

    """
    def __init__(self, connection):
        self.connection = connection
        self.channel = None
        self.logger = logging.getLogger(__name__)

    def set_logger(self, logger) -> None:
        """
        Set logger for class.

        :param logger: Logger to be used
        :return: None
        """
        self.logger = logger

    async def connect(self, channel_id: Optional[int] = None) -> None:
        """
        Establish a connection (channel) to RabbitMQ.

        :param channel_id: Channel id to use
        :return: None
        :raises ConnectionError: If connection not established
        """
        if self.connection is None or self.connection.is_closed:
            raise ConnectionError("Connection not established")
        self.channel = await self.connection.channel(channel_id)

    async def disconnect(self) -> None:
        """
        Disconnect from RabbitMQ.

        :return: None
        """
        if self.channel is not None and not self.channel.is_closed:
            await self.channel.close()


class Publisher(ConnectionSetup):
    """
    Class to publish messages to RabbitMQ.

    Note: When `exchange != ""` the publisher will declare the exchange if it does not exist.

    :param connection: Connection to RabbitMQ
    :param exchange: Name of exchange to publish messages to

    Example::
        >>> import aio_pika
        >>> from easybits_module_interface import Publisher
        >>> from easybits_module_interface.models import Message
        >>> connection = await aio_pika.connect("amqp://localhost/", loop=loop)
        >>> publisher = Publisher(connection, "test_exchange")
        >>> await publisher.publish(Message.from_dict(**message_dict))
        >>> await publisher.disconnect()
    """
    def __init__(self, connection, exchange):
        super().__init__(connection)
        self.exchange_name = exchange
        self.default_exchange = None

    async def connect(self):
        """
        Establish a connection to RabbitMQ.

        :return: None
        """
        await super().connect()
        if self.exchange_name == "":
            self.logger.warning("No exchange name provided, using default exchange")
            self.default_exchange = self.channel.default_exchange
            return
        # ensure default exchange exists
        self.default_exchange = await self.channel.declare_exchange(
            self.exchange_name, ExchangeType.FANOUT, durable=True
        )

    async def publish(self, message: Message, exchange: Optional[str] = None, queue: str = ""):
        """
        Publish a message to RabbitMQ.

        Internally, establishes a connection via a new channel, and publishes the message to the desired target.
        Converts :py:obj:`~message` to a dictionary, then to a JSON string and finally to a byte string.
        If :py:obj:`~exchange` is empty, the message will be published to the Publisher's default exchange.

        **Note**: <u>Please do not publish messages to the default exchange & default queue,
                  always provide a queue or exchange name.</u>

        Values for :py:obj:`~exchange`:
            - None: Use the default exchange of the module
            - "": Use the RMQ default exchange
            - "exchange_name": Use the exchange with the given name

        :param message: Message to be published
        :param exchange: Name of exchange to publish message to
        :param queue: Name of queue to publish message to
        :return: None
        """
        if self.channel is None or self.channel.is_closed:
            await self.connect()

        if exchange is None:
            ex = self.default_exchange
        elif exchange == "":
            ex = self.channel.default_exchange
        else:
            ex = await self.channel.get_exchange(exchange)

        msg_body = json.dumps(asdict(message)).encode('utf-8')
        pika_msg = PikaMessage(body=msg_body)
        await ex.publish(
            pika_msg,
            routing_key=queue
        )


class Consumer(ConnectionSetup):
    """
    Class to consume messages from RabbitMQ.

    :param connection: Connection to RabbitMQ

    Example::
        >>> import aio_pika
        >>> from easybits_module_interface import Consumer
        >>> from easybits_module_interface.models import Message
        >>> connection = await aio_pika.connect("amqp://localhost/", loop=loop)
        >>> consumer = Consumer(connection)
        >>> consumer.add_callback("test_queue", callback)
        >>> await consumer.run()
        >>> await consumer.disconnect()
    """
    def __init__(self, connection):
        super().__init__(connection)
        self.queue = None
        self.callback_map = {}

    @staticmethod
    def __callback_wrapper(fun):
        """
        Wraps input :py:obj:`~fun` (callbacks) so that signature matches :py:obj:`~pika`s consumption call.
        Internally builds :py:obj:`~models.Message` from incoming message body, and passes it to :py:obj:`~fun`.

        This way the callback can use Message class to interact with the message instead of handling a dictionary.

        :param fun: Function to be decorated
        :return: wrapped :py:obj:`~fun`
        """
        async def inner(msg: AbstractIncomingMessage):
            message = Message.from_dict(**json.loads(msg.body.decode('utf-8')))
            await fun(message)
            await msg.ack()

        return inner

    def add_callback(self, queue: str, callback: Callable[[Message], None], exchange: str = None, is_durable: bool = True):
        """
        Add a callback to be called when a message is received on :py:obj:`~queue`.

        :param queue: Name of queue to consume messages from
        :param callback: Callback function
        :param exchange: Name of exchange to bind queue to
        :param is_durable: Whether to create a durable queue or not
        """
        self.callback_map[queue] = (self.__callback_wrapper(callback), exchange, is_durable)

    async def run(self):
        """
        Run the consumer.

        This will start the consumer, i.e. register configured callbacks under the
        desired queues.
        In case the startup process fails, the consumer will be closed
        and the occurred exception will be raised.

        **Note:** Register callbacks using :py:obj:`~add_callback` before calling this method.
        """
        try:
            await self.connect()
            await self.channel.set_qos(prefetch_count=1)
            for queue_name, (callback, exchange_name, is_durable) in self.callback_map.items():
                queue = await self.channel.declare_queue(queue_name, durable=is_durable)
                if exchange_name is not None:
                    exchange = await self.channel.declare_exchange(
                        exchange_name, ExchangeType.FANOUT, durable=True
                    )
                    await queue.bind(exchange, queue_name)
                await queue.consume(callback, no_ack=False)

            self.logger.info("LISTENING TO {}".format(' '.join(self.callback_map.keys())))
            await asyncio.Future()
        except Exception as e:
            self.logger.exception("Exception in consumer")
            raise e
        finally:
            await self.disconnect()


class ModuleInterface:
    """
    Base class for module implementations.

    :param connection: Connection to RabbitMQ
    :param exchange: Name of exchange to publish messages to
    :param default_consumer_queue: Name of queue to consume messages from

    Example:
        >>> import aio_pika
        >>> import asyncio
        >>> from easybits_module_interface import ModuleInterface
        >>> from easybits_module_interface.models import Message
        >>> class MyModule(ModuleInterface):
        ...     async def callback(self, message: Message):
        ...         print(message)
        >>> connection = await aio_pika.connect("amqp://localhost/", loop=loop)
        >>> module = MyModule(connection, "test_exchange", "test_queue")
        >>> await module.run()
        >>> try:
        >>>     await asyncio.Future()
        >>> finally:
        >>>     await module.shutdown()

    Or as a `system_message` only module:
        >>> class MyModule(ModuleInterface):
        ...     def __init__(self, connection):
        ...         super().__init__(connection, "", None)
        ...         self.add_callback("system_messages_my_module", self.callback)
        ...
        ...     async def callback(self, message: Message):
        ...         print(message)
        ...
        >>> connection = await aio_pika.connect("amqp://localhost/", loop=loop)
        >>> module = MyModule(connection)
    """
    def __init__(
            self, connection, exchange: Optional[str] = None,
            default_consumer_queue: Optional[str] = None
    ) -> None:
        #: Provide placeholder for logger
        self.logger = None

        #: default connection to message broker
        self.__connection = connection

        #: consumer entity to consume incoming messages
        self.__consumer = Consumer(self.__connection)

        if default_consumer_queue is not None:
            #: add default callback for default queue
            self.__consumer.add_callback(default_consumer_queue, self.callback)

        #: publisher entity to publish outgoing messages
        self.__publisher = Publisher(self.__connection, exchange)

        #: initializing the default logger
        self.init_logger('module_interface')

    def __new__(cls, *args, **kwargs):
        # guard against instantiating the base class
        if cls is ModuleInterface:
            raise TypeError(
                f"only children of '{cls.__name__}' may be instantiated"
            )
        return super().__new__(cls)

    def init_logger(self, name: str) -> None:
        """
        Initialize logger for module implementation.

        :param name: Name of logger
        :return: None
        """
        name = f'{name.lower()}.interface'

        self.logger = logging.getLogger(name)
        self.__consumer.set_logger(self.logger)
        self.__publisher.set_logger(self.logger)

    def capture_exception(self, exception) -> None:
        """
        Helper to capture exceptions during message processing.

        Logs exception to:
        - stdout logs
        - sentry

        :param exception: The exception to capture
        :returns: Nothing
        """
        self.logger.error(traceback.format_exc())
        sentry_sdk.capture_exception(exception)

    def get_loggable_message(self, message: Message, exclude_outgoing: bool = False) -> dict:
        """
        Helper method to consistently log messages

        :param message: The message to convert
        :param exclude_outgoing: Indicates if outgoing message should be masked, too
        :returns: masked message dictionary
        """
        msg = asdict(message)
        keys_to_hide = {
            'message.incoming_file': 'FILE_PLACEHOLDER',
        }
        if exclude_outgoing:
            keys_to_hide['message.outgoing'] = 'OUTGOING_PLACEHOLDER'
        for key, placeholder in keys_to_hide.items():
            elems = key.split('.')
            ref = msg
            for idx, e in enumerate(elems):
                if ref is None or e not in ref:
                    break
                if idx < len(elems) - 1:
                    ref = ref[e]
                else:
                    ref[e] = placeholder
        return msg

    def add_callback(self, queue: str, callback, exchange=None, is_durable=True) -> None:
        """
        Add a callback to be called when a message is received on :py:obj:`~queue`.

        :param queue: Name of queue to consume messages from
        :param callback: Callback function
        :param exchange: Name of exchange to bind queue to
        :param is_durable: Whether to create a durable queue or not
        :return: None
        """
        self.__consumer.add_callback(queue, callback, exchange, is_durable)

    async def run(self) -> None:
        """
        Run the module.

        This will start the consumer and wait for it to finish.
        """
        await self.__publisher.connect()
        await self.__consumer.run()

    async def shutdown(self) -> None:
        """
        Shutdown the module.
        """
        await self.__consumer.disconnect()
        await self.__publisher.disconnect()
        await self.__connection.close()
        self.logger.info("Shutting down")

    async def publish(self, message: Message, exchange: str = None, queue: str = "") -> None:
        """
        Publish a message to RabbitMQ.

        If :py:obj:`~exchange` is empty, the message will be published to the Publisher's default exchange.
        If :py:obj:`~queue` is empty, the message will be published to the RMQ default queue.

        :param message: Message to be published
        :param exchange: Name of exchange to publish message to
        :param queue: Name of queue to publish message to
        :return: None
        """
        await self.__publisher.publish(message, exchange, queue)

    async def callback(self, message: Message) -> None:
        """
        Callback function to be implemented by module implementations.

        :param message: Message received from consumer queue
        :return: None
        """
        raise NotImplementedError("Callback not implemented")


async def bootstrap_module_class(module_implementation, **config) -> None:
    """
    Main entry point for module implementations, aka application loop.

    :param module_implementation: ModuleInterface implementation (as class)
    :param config: Configuration dictionary
    :return: None

    ================  =============  =============
    Key               Type           Description
    ================  =============  =============
    logger            dict           Logger configuration
    logger.level      str            Logger level
    logger.path       str            Path to logger configuration file
    logger.name       str            Name of logger
    ================= =============  =============
    broker            dict           Broker configuration
    broker.uri        str            Broker uri
    """
    logger_config = config.get('logger', None)
    broker_config = config.get('broker', None)

    assert 'uri' in broker_config, 'broker uri required!'

    logger_name = None
    if logger_config is not None:
        logging.basicConfig()
        if 'level' in logger_config:
            logging.getLogger().setLevel(logger_config['level'])

        if 'path' in logger_config:
            # set up logger from file
            logging.config.fileConfig(logger_config['path'])

        # create logger
        logger_name = logger_config.get('name', 'module_interface')
        _ = logging.getLogger(name=logger_name)

    connection = await connect_robust(broker_config['uri'])
    module = module_implementation(connection)
    module.init_logger(logger_name)
    return module


async def main(module_implementation, **config) -> None:
    """
    Main entry point for module implementations, aka application loop.

    :param module_implementation: ModuleInterface implementation (as class)
    :param config: Configuration dictionary
    :return: None

    ================  =============  =============
    Key               Type           Description
    ================  =============  =============
    logger            dict           Logger configuration
    logger.level      str            Logger level
    logger.path       str            Path to logger configuration file
    logger.name       str            Name of logger
    ================= =============  =============
    broker            dict           Broker configuration
    broker.uri        str            Broker uri
    """
    module = await bootstrap_module_class(module_implementation, **config)
    try:
        await module.run()
    except Exception as e:
        module.logger.error(e)
    finally:
        module.logger.info("MAIN Shutting down")
        await module.shutdown()



import logging
from dataclasses import dataclass
from typing import Any, Callable, List, Optional, Union

logger = logging.getLogger()


@dataclass
class Transition:
    """Data class to store information about a transition.

    """
    #: The original state of the transition
    from_state: str
    #: The target state of the transition.
    to_state: str
    #: Validator for incoming message, returns True if message is valid, else False
    validator: Union[bool, Callable[[Any], bool]]
    #: Callback to call in case the transition is successful
    success: Optional[Callable[[Any], None]] = None
    #: Callback to call in case the transition is unsuccessful
    failure: Optional[Callable[[Any], None]] = None


@dataclass
class TransitionResult:
    """Data class representation of a transition result."""
    #: The side-effect action that can be triggered for this resulting transition
    action: Optional[Callable[[Any], None]]
    #: The resulting state of a transition
    state: Optional[str]

    def __call__(self, *args, **kwargs):
        if self.action is not None:
            return self.action(*args, **kwargs)


class IllegalTransitionError(Exception):
    """Exception class to signalize adding an illegal transition to a state machine"""


class StateMachine:
    """Implementation of directed graph for state machine.

    Example:
        >>> machine = StateMachine()
        >>> transition = Transition(
        ...     from_state='A',
        ...     to_state='B',
        ...     validator=lambda x: 'foo' in x,
        ...     success=lambda: 0,
        ...     failure=lambda: -1
        ... )
        >>> machine.define_transition(transition)
        >>> res = machine.make_transition('A', 'my message foo')
        >>> assert res.state == 'B'
        >>> assert res.action() == 0
    """
    def __init__(self):
        self._transitions: List[Transition] = []

    def _validate_transition(self, transition: Transition) -> None:
        """Validation method to add a new transition into the state machine.

        :param transition: The transition to add to the state machine
        :raises IllegalTransitionError: in case there is already a transition for
                                        the provided states registered.
        """
        for existing_transition in self._transitions:
            is_illegal_transition = (
                existing_transition.from_state == transition.from_state
                and existing_transition.to_state == transition.to_state
            )
            if is_illegal_transition:
                raise IllegalTransitionError(
                    f'There is already a transition existing with'
                    f'from_state={transition.from_state} and '
                    f'to_state={transition.to_state}'
                )

    def define_transition(self, transition: Transition) -> None:
        """Method to register transitions between states.

        :param transition: The transition to add
        """
        self._validate_transition(transition)
        self._transitions.append(transition)

    def make_transition(self, current_state: str, payload: str) -> Optional[TransitionResult]:
        """Method to perform transitions

        :param current_state: the current state inside the state machine.
        :param payload: The payload to validate the transition on.
        :returns: A :py:obj:`~.TransitionResult` if a transition is available else None/
        """
        transition = list(filter(lambda x: x.from_state == current_state, self._transitions))
        if not transition:
            logger.error(f"No transition defined from {current_state}")
            return

        transition = transition[0]
        is_success = transition.validator

        if callable(transition.validator):
            is_success = transition.validator(payload)

        if is_success:
            return TransitionResult(action=transition.success, state=transition.to_state)
        return TransitionResult(action=transition.failure, state=None)
